/**
 * 
 */
package de.mpc.prolic.comparison;

import java.util.HashMap;
import java.util.ResourceBundle;

import de.mpc.prolic.exceptions.EmptySequenceException;
import de.mpc.prolic.exceptions.NoPercentIdentityValueComputedException;
import de.mpc.prolic.misc.Tools;

import org.apache.logging.log4j.Logger;
import org.biojava.nbio.alignment.NeedlemanWunsch;
import org.biojava.nbio.alignment.SimpleGapPenalty;
import org.biojava.nbio.alignment.SimpleSequencePair;
import org.biojava.nbio.alignment.SubstitutionMatrixHelper;
import org.biojava.nbio.alignment.template.SubstitutionMatrix;
import org.biojava.nbio.core.exceptions.CompoundNotFoundException;
import org.biojava.nbio.core.sequence.ProteinSequence;
import org.biojava.nbio.core.sequence.compound.AminoAcidCompound;

/**
 * This class provides different static methods for comparison of two protein sequences.
 * @author Michael Kohl
 *
 */
public class SequenceBasedComparison {
/*******************************************************************************
* Constants 
******************************************************************************/

	private final String k_pid="pid";							//Key for Percent identity calculated according to the choosen PID method.
	private final String k_similarity="similarity";				//Key for the calculated NW score given as a similarity value between 0.0 and 1.0 according to the Needleman-Wunsch implementation of BioJava. 
																//This equals (getScore() - getMinScore()) / (getMaxScore() - getMinScore()).
	private final String k_similarMatch="similarMatch"; 		//Key for the number of positions with similar AA on both aligned protein sequences					
	private final String k_identMatch="identMatch";				//Key for the number of identical positions of aligned sequences 
	private final String k_AlignedSeq="AlSeq";					//Key for the aligned sequence (for the query sequence (should be identical for the target as well).	
	private final String k_AlignedSeqLength="AlSeqLength";		//Key for the length of the aligned sequence (for the query sequence (should be identical for the target as well).
	private final String k_score="score";						//Key for the score calculated by the alignment algorithm
	private final String k_resString="resString";				//Key for the basic string based graphical representation of the pairwise comparison
	private final String k_numbGaps="numbGaps";					//Key for for the absolute value of gaps calculated by the alignment algorithm
	private final String k_numbIntGaps="numbIntGaps";			//Key for the number of internal gaps 
	private final String k_numbTermGaps="numbTermGaps";			//Key for the number of terminal gaps (C and N terminal)
	private final String defaultMatrix = "BLOSUM62";			//Default substitution matrix
//Messages from the ResourceBundle
	private final String k_compoundNotFound = "m_compoundNotFound";
	private final String k_noPIDcomputed = "m_computationPIDfailes";
	private final String spacer = " ";
//Switch for the debug mode
	private final boolean debugMode=false;

/*******************************************************************************
* Member
******************************************************************************/
	private String pidMethod;
	private String substMatrixID;
	private int gapOpen;
	private int gapExt;
	private Logger log4j;
	private ResourceBundle bundleMess = ResourceBundle.getBundle("de.mpc.prolic.resources.MessageBundle");


	/**
	 * Empty constructor
	 */
	public SequenceBasedComparison() {

	}
	
	/**
	 * Enables logging capabilites along with creation of a SequenceBasedComparison instance
	 */
	public SequenceBasedComparison(Logger log4j_) {
		log4j = log4j_;
	}
	
	/**
	 * The method performs a pairwise sequence based comparison.
	 * @param queryAsString
	 * @param targetAsString
	 * @return a map with the results of a pairwise sequence based comparison of two protein sequences.
	 * @throws EmptySequenceException 
	 */
	public HashMap<String,String> pairwiseSequenceComp(String queryAsString, String targetAsString) throws EmptySequenceException, CompoundNotFoundException, NoPercentIdentityValueComputedException{
		HashMap <String,String> results = new HashMap<String,String>();
		double pid=-9;
		double identMatch;
		double numbGaps=0;
		double numbTerminalGaps=0;
		double numbInternalGaps=0;
		double lengthAlignedSeq;
		double meanSeqLength;
		ProteinSequence query = getProtSeqQuery(queryAsString); 
		ProteinSequence target = this.getProtSeqTarget(targetAsString);
		if ((query.getLength()==0) || (target.getLength()==0)){//Comparison of empty sequences raises an exception.
			throw new EmptySequenceException();			
		}
		meanSeqLength=(query.getLength()+target.getLength())/2.0;
		SimpleGapPenalty gp = new SimpleGapPenalty(gapOpen, gapExt);
		SubstitutionMatrix<AminoAcidCompound> sm= getSubstMatr(substMatrixID);
		NeedlemanWunsch <ProteinSequence, AminoAcidCompound> nw= new NeedlemanWunsch <ProteinSequence, AminoAcidCompound>(query, target, gp, sm);
		SimpleSequencePair <ProteinSequence, AminoAcidCompound> p = (SimpleSequencePair <ProteinSequence, AminoAcidCompound>)nw.getProfile(); 
		identMatch=p.getNumIdenticals();	
		/* 
		 * System.out.println();
		 * this gives the score matrix:
		 * System.out.println(nw.getScoreMatrixAsString());
		 */
		lengthAlignedSeq=p.getAlignedSequence(query).getLength();
		for (int i=1; i<lengthAlignedSeq+1; i++){
			if(p.hasGap(i)){
				numbGaps++;					
			}
		}
		if(debugMode){//Aligned Sequence: Results for the target
			System.out.println(p.getAlignedSequence(target).getLength());
			System.out.println(p.getAlignedSequence(target));
		}
		numbTerminalGaps=Tools.calculateExternalGaps(p.toString());
		numbInternalGaps=numbGaps-numbTerminalGaps;
		//Calculating the PID
		if(pidMethod.equals("PID1")){//Denominator: Length of the shorter sequence
		   	  if (query.getLength()<=target.getLength()){
		   		  	pid=100*identMatch/query.getLength();						
		   	  }else{
					pid=100*identMatch/target.getLength();
		   	  }
		}	
		if(pidMethod.equals("PID2")){//Denominator: number aligned positions=length of the aligned sequence (without gaps)
		   	  pid=100*identMatch/lengthAlignedSeq;
		}
		if(pidMethod.equals("PID3")){//Denominator: mean sequence length //equation according to May AC: Percent sequence identity; the need to be explicit. Structure (Camb) 2004, 12(5):737-8. [09692126 (Print) Letter]
			pid=100*identMatch/meanSeqLength;  
			if(debugMode){
		   		System.out.println("Length query sequence: "+query.getLength());
		   		System.out.println("Length target sequence: "+target.getLength());
		   		System.out.println("Mean sequence length: "+meanSeqLength);
		   	}
		}
		if(pidMethod.equals("PID4")){//Denominator: number aligned positions+ number internal gaps 
		   	  pid=100*identMatch/(lengthAlignedSeq+numbInternalGaps);
		}
		if(pid==-9){ 
		   	throw new NoPercentIdentityValueComputedException(bundleMess.getString(k_noPIDcomputed)+spacer);
		}	
		//Storing the comparison results
		results.put(k_score, String.valueOf(nw.getScore()));
		results.put(k_identMatch, String.valueOf(identMatch));
		results.put(k_AlignedSeqLength, String.valueOf(p.getAlignedSequence(query).getLength()));	
		results.put(k_AlignedSeq, String.valueOf(p.getAlignedSequence(query).toString()));				
		results.put(k_similarMatch, String.valueOf(p.getNumSimilars()));
		results.put(k_similarity, String.valueOf(nw.getSimilarity()));
		results.put(k_pid, String.valueOf(pid));
		results.put(k_resString, p.toString());		
		results.put(k_numbGaps, String.valueOf(numbGaps));
		results.put(k_numbIntGaps, String.valueOf(numbInternalGaps));
		results.put(k_numbTermGaps, String.valueOf(numbTerminalGaps));			
		return results;
	}
	
	/**
	 * Method is necessary for determination that the query sequence causes the exception
	 * @param queryAsString
	 * @return the query protein sequence.
	 * @throws CompoundNotFoundException 
	 */
	private ProteinSequence getProtSeqQuery(String queryAsString) throws CompoundNotFoundException{
		ProteinSequence query=null;
		try {
			query = new ProteinSequence(queryAsString);			
		} catch (CompoundNotFoundException e) {
			throw new CompoundNotFoundException(bundleMess.getString(k_compoundNotFound)+spacer+queryAsString);
		}
		return query;
	}
	
	/**
	 * Method is necessary for determination that the target sequence causes the exception.
	 * @param targetAsString
	 * @return the target protein sequence.
	 * @throws CompoundNotFoundException
	 */
	public ProteinSequence getProtSeqTarget(String targetAsString) throws CompoundNotFoundException{ 
		ProteinSequence target=null;
		try {
			target = new ProteinSequence(targetAsString);			
		} catch (CompoundNotFoundException e) {
			throw new CompoundNotFoundException(bundleMess.getString(k_compoundNotFound)+spacer+targetAsString);
		}
		return target;
	}
	
	/**
	 * Auxiliary method implemented for the sake of clarity. Returns a substitution matrix according to the specified
	 * textual identifier.
	 * @param substMatrixID
	 * @return a substitution matrix generated with respect to the parameter specified in the ProLiC configuration file.
	 */
	private SubstitutionMatrix<AminoAcidCompound> getSubstMatr(String substMatrixID){
		SubstitutionMatrix<AminoAcidCompound> sm=null;
		boolean idIsValid=false;
		String mess;
		if (substMatrixID.equalsIgnoreCase("blosum30")){
			sm=SubstitutionMatrixHelper.getBlosum30();
			idIsValid=true;
		}
		if (substMatrixID.equalsIgnoreCase("blosum35")){
			sm=SubstitutionMatrixHelper.getBlosum35();
			idIsValid=true;
		}
		if (substMatrixID.equalsIgnoreCase("blosum40")){
			sm=SubstitutionMatrixHelper.getBlosum40();
			idIsValid=true;
		}
		if (substMatrixID.equalsIgnoreCase("blosum45")){
			sm=SubstitutionMatrixHelper.getBlosum45();
			idIsValid=true;
		}
		if (substMatrixID.equalsIgnoreCase("blosum50")){
			sm=SubstitutionMatrixHelper.getBlosum50();
			idIsValid=true;
		}
		if (substMatrixID.equalsIgnoreCase("blosum55")){
			sm=SubstitutionMatrixHelper.getBlosum55();
			idIsValid=true;
		}
		if (substMatrixID.equalsIgnoreCase("blosum60")){
			sm=SubstitutionMatrixHelper.getBlosum60();
			idIsValid=true;
		}
		if (substMatrixID.equalsIgnoreCase("blosum62")){
			sm=SubstitutionMatrixHelper.getBlosum62();
			idIsValid=true;
		}
		if (substMatrixID.equalsIgnoreCase("blosum65")){
			sm=SubstitutionMatrixHelper.getBlosum65();
			idIsValid=true;
		}
		if (substMatrixID.equalsIgnoreCase("blosum70")){
			sm=SubstitutionMatrixHelper.getBlosum70();
			idIsValid=true;
		}
		if (substMatrixID.equalsIgnoreCase("blosum75")){
			sm=SubstitutionMatrixHelper.getBlosum75();
			idIsValid=true;
		}
		if (substMatrixID.equalsIgnoreCase("blosum80")){
			sm=SubstitutionMatrixHelper.getBlosum80();
			idIsValid=true;
		}
		if (substMatrixID.equalsIgnoreCase("blosum85")){
			sm=SubstitutionMatrixHelper.getBlosum85();
			idIsValid=true;
		}
		if (substMatrixID.equalsIgnoreCase("blosum90")){
			sm=SubstitutionMatrixHelper.getBlosum90();
			idIsValid=true;
		}
		if (substMatrixID.equalsIgnoreCase("blosum100")){
			sm=SubstitutionMatrixHelper.getBlosum100();
			idIsValid=true;
		}
		if (substMatrixID.equalsIgnoreCase("pfam250")){ //Returns PAM 250 matrix by Dayhoff
			sm=SubstitutionMatrixHelper.getPAM250();
			idIsValid=true;
		}
		if (!idIsValid){
			sm=SubstitutionMatrixHelper.getBlosum62();//default substitution matrix
			mess = bundleMess.getString("m_comparison_seq_default_subsMatrix_info")+spacer+defaultMatrix;
			log4j.info(mess);
		}
		return sm;		
	}
	
/*******************************************************************************
* Setter and getter
******************************************************************************/
	public void setPidMethod(String pidMethod_){
		this.pidMethod=pidMethod_;
	}
	
	public String getPidMethod(){
		return this.pidMethod;
	}
	
	public void setSubstMatrixID(String substMatrixID_){
		this.substMatrixID=substMatrixID_;
	}
	
	public String getSubstMatrixID(){
		return this.substMatrixID;
	}
	
	public void setGapOpen(int gapOpen_){
		this.gapOpen=gapOpen_;
	}
	
	public int getGapOpen(){
		return this.gapOpen;
	}
	
	public void setGapExt(int gapExt_){
		this.gapExt=gapExt_;
	}
	
	public int getGapExt(){
		return this.gapExt;
	}
	
	public void setLog4j(Logger log4j_){
		this.log4j=log4j_;
	}
}
