/**
 * 
 */
package de.mpc.prolic.model;

/**
 * The superclass for all biological molecules that contain sequence information. Here a biomolecule is either a protein 
 * or some kind of DNA or RNA molecule, but not another sort of biological molecule like e.g. a carbohydrate.
 * This class is implemented with respect to future versions of ProLiC that will comprise comparisons with further biological molecules.  
 * @author Michael Kohl
 */
public class Biomolecule implements Comparable<Object> {
	
	/*******************************************************************************
	 * Members
	 ******************************************************************************/		
	private String identifier; 			//The identifier string for this biomolecule
	private String uniqueIdentifier;	/*Unique identifier of this biomolecule, i.e. a combination of the identifier and the origin. Because the identifier is
										unique within a given original list, the combination is unique with respect to the current ProLiC comparison.*/
	private String biomoleculeType;		//Specifies the type of the biomolecule for a specific instance.
	private String sequence;			//Either an amino acid sequence (proteins) or a nucleotide sequence (DNA/RNA)
	private String origin; 				/*Each protein or gene has a reference to its origin, i.e the experiment/biomolecule list where the current 
										biomolecule was identified or measured. Here the origin is equivalent with the input file name that contains
										the biomolecule information used for creation of this instance of biomolecule.*/
	private double regulation;			/*Each biomolecule can store information about its regulation, which is calculated from a comparison of 
										some kind of metrics between different conditions of this biomolecule*/
	/*******************************************************************************
	 * Constructors
	 ******************************************************************************/
	/**
	 * Empty constructor
	 */
	public Biomolecule() {

	}
	
	/**
	 * For biomolecules that are part of higher level entities (e.g. peptides that are associated with their bearing proteins)
	 * Here, the information about the origin is stored as part of the higher level entity.
	 * @param sequence_
	 */
	public Biomolecule(String sequence_){
		sequence=sequence_;
	}

	/**
	 * The default constructor for generating instances of Biomolecule
	 * @param sequence_
	 */
	public Biomolecule(String sequence_, String origin_) {
		sequence=sequence_;
		origin=origin_;
	}
	

	@Override
	/**
	 * Supports the comparison of two biomolecules with respect to their unique identifier, which is a combination of an identifier and the
	 * origin (i.e. the file name where the biomolecule was originally listed). 
	 */
	public int compareTo(Object o) {
		Biomolecule b = (Biomolecule) o;
		return this.getUniqueIdentifier().compareTo(b.getUniqueIdentifier());
	}
	
	/*******************************************************************************
	 * Setter and Getter
	 ******************************************************************************/
	public String getSequence(){
		return this.sequence;
	}
	
	public void setSequence(String sequence_){
		this.sequence=sequence_;
	}
	
	public String getIdentifier(){
		return this.identifier;
	}
	
	public void setIdentifier(String identifier_){
		this.identifier=identifier_;
	}
	
	public String getOrigin(){
		return origin;
	}
	
	public void setOrigin(String origin_){
		origin = origin_;
	}
	
	public String getUniqueIdentifier(){
		return uniqueIdentifier;
	}
	
	public void setUniqueIdentifier(String uniqueIdentifier_){
		uniqueIdentifier = uniqueIdentifier_;
	}
	
	public String getBiomoleculeType(){
		return biomoleculeType;
	}
	
	public void setBiomoleculeType(String biomoleculeType_){
		biomoleculeType = biomoleculeType_;
	}	
	
	public double getRegulation(){
		return this.regulation;
	}
	
	public void setRegulation(double regulation_){
		regulation = regulation_;
	}	
}
